@@ -3,23 +3,22 @@ import decimal
 import enum
 import fractions
 import json
+import os
 import pathlib
 import tempfile
 import textwrap
 
 import pytest
-import hypothesis as hyp
 import numpy as np
 
 import demes
-import demes.hypothesis_strategies
 import tests
 
 
 def jacobs_papuans():
     """
     XXX: This model is for testing only and has not been vetted for accuracy!
-         Use examples/jacobs_papuans.yml, or the PapuansOutOfAfrica_10J19 model
+         Use examples/jacobs_papuans.yaml, or the PapuansOutOfAfrica_10J19 model
          from stdpopsim instead.
     """
     generation_time = 29
@@ -140,25 +139,25 @@ def jacobs_papuans():
         demes=["Ghost3", "Eurasia"], rate=4.42e-4, start_time=T_Eu_bottleneck
     )
 
-    b.add_pulse(source="NeaI", dest="EastAsia", proportion=0.002, time=883)
-    b.add_pulse(source="NeaI", dest="Papua", proportion=0.002, time=1412)
-    b.add_pulse(source="NeaI", dest="Eurasia", proportion=0.011, time=1566)
-    b.add_pulse(source="NeaI", dest="Ghost1", proportion=0.024, time=1853)
+    b.add_pulse(sources=["NeaI"], dest="EastAsia", proportions=[0.002], time=883)
+    b.add_pulse(sources=["NeaI"], dest="Papua", proportions=[0.002], time=1412)
+    b.add_pulse(sources=["NeaI"], dest="Eurasia", proportions=[0.011], time=1566)
+    b.add_pulse(sources=["NeaI"], dest="Ghost1", proportions=[0.024], time=1853)
 
     m_Den_Papuan = 0.04
     p = 0.55  # S10.i p. 31
     T_Den1_Papuan_mig = 29.8e3 / generation_time
     T_Den2_Papuan_mig = 45.7e3 / generation_time
     b.add_pulse(
-        source="DenI1",
+        sources=["DenI1"],
         dest="Papua",
-        proportion=p * m_Den_Papuan,
+        proportions=[p * m_Den_Papuan],
         time=T_Den1_Papuan_mig,
     )
     b.add_pulse(
-        source="DenI2",
+        sources=["DenI2"],
         dest="Papua",
-        proportion=(1 - p) * m_Den_Papuan,
+        proportions=[(1 - p) * m_Den_Papuan],
         time=T_Den2_Papuan_mig,
     )
 
@@ -167,8 +166,7 @@ def jacobs_papuans():
 
 class TestLoadAndDump:
     def test_bad_format_param(self):
-        examples_path = pathlib.Path(__file__).parent.parent / "examples"
-        ex = next(examples_path.glob("*.yml"))
+        ex = tests.example_files()[0]
         with open(ex) as f:
             ex_string = f.read()
 
@@ -275,8 +273,8 @@ class TestLoadAndDump:
             assert f"{deme.name}" in string
         assert "pulses" in string
         for pulse in g.pulses:
-            assert "source" in string
-            assert pulse.source in string
+            assert "sources" in string
+            assert pulse.sources[0] in string
             assert "dest" in string
             assert pulse.dest in string
         assert "migrations" in string
@@ -295,14 +293,14 @@ class TestLoadAndDump:
         g = jacobs_papuans()
         dumps_str = demes.dumps(g, format=format, simplified=simplified)
         with tempfile.TemporaryDirectory() as tmpdir:
-            tmpfile1 = pathlib.Path(tmpdir) / "temp1.yml"
+            tmpfile1 = pathlib.Path(tmpdir) / "temp1.yaml"
             # tmpfile is os.PathLike
             demes.dump(g, tmpfile1, format=format, simplified=simplified)
             with open(tmpfile1) as f:
                 yaml_str1 = f.read()
             assert yaml_str1 == dumps_str
 
-            tmpfile2 = pathlib.Path(tmpdir) / "temp2.yml"
+            tmpfile2 = pathlib.Path(tmpdir) / "temp2.yaml"
             # tmpfile is str
             demes.dump(g, str(tmpfile2), format=format, simplified=simplified)
             with open(tmpfile2) as f:
@@ -442,7 +440,7 @@ class TestLoadAndDump:
     def test_examples_load_dump_load(self, yaml_file, format, simplified):
         g1 = demes.load(yaml_file, format="yaml")
         with tempfile.TemporaryDirectory() as tmpdir:
-            tmpfile = pathlib.Path(tmpdir) / "temp.yml"
+            tmpfile = pathlib.Path(tmpdir) / "temp.yaml"
             # dump and load files
             demes.dump(g1, tmpfile, format=format, simplified=simplified)
             g2 = demes.load(tmpfile, format=format)
@@ -490,10 +488,6 @@ class TestLoadAndDump:
                     if format == "yaml":
                         self.check_yaml_output_is_pretty(g, tmpfile, simplified)
 
-    @hyp.given(g=demes.hypothesis_strategies.graphs())
-    def test_dump_load(self, g):
-        self.check_dump_load_roundtrip(g)
-
     def test_int_subclass(self):
         # Check that subclasses of int are round-trippable.
         class Ne(enum.IntEnum):
@@ -565,10 +559,10 @@ class TestLoadAndDump:
         self.check_dump_load_roundtrip(b.resolve())
 
         b.add_pulse(
-            source="A",
+            sources=["A"],
             dest="B",
             time=T[1],
-            proportion=decimal.Decimal("0.0022"),
+            proportions=[decimal.Decimal("0.0022")],
         )
         self.check_dump_load_roundtrip(b.resolve())
 
@@ -594,3 +588,217 @@ class TestLoadAndDump:
         assert data["migrations"][0]["start_time"] == "Infinity"
         g2 = demes.loads(json_str, format="json")
         g2.assert_close(g1)
+
+    def test_load_with_null_deme_start_time(self):
+        model = """
+time_units: generations
+demes:
+- name: a
+  start_time: null
+  epochs:
+  - {end_time: 0, start_size: 1}
+"""
+        with pytest.raises(ValueError, match="must have a non-null value"):
+            demes.loads(model)
+
+    def test_load_with_null_deme_name(self):
+        model = """
+time_units: generations
+demes:
+- name: null
+  epochs:
+  - {end_time: 0, start_size: 1}
+"""
+        with pytest.raises(ValueError, match="must have a non-null value"):
+            demes.loads(model)
+
+    def test_load_with_null_time_unit(self):
+        model = """
+time_units: null
+demes:
+- name: a
+  epochs:
+  - {end_time: 0, start_size: 1}
+"""
+        with pytest.raises(ValueError, match="must have a non-null value"):
+            demes.loads(model)
+
+    def test_load_with_null_epochs(self):
+        model = """
+time_units: generations
+demes:
+- name: a
+  epochs:
+"""
+        with pytest.raises(ValueError, match="must have a non-null value"):
+            demes.loads(model)
+
+    def test_load_with_null_epoch_end_time(self):
+        model = """
+time_units: generations
+demes:
+- name: a
+  epochs:
+  - {end_time: null, start_size: 1}
+"""
+        with pytest.raises(ValueError, match="must have a non-null value"):
+            demes.loads(model)
+
+    def test_load_with_null_migrations(self):
+        model = """
+time_units: generations
+demes:
+- name: a
+  epochs:
+  - {start_size: 1}
+- name: b
+  epochs:
+  - {start_size: 1}
+migrations:
+"""
+        with pytest.raises(ValueError, match="must have a non-null value"):
+            demes.loads(model)
+
+    def test_load_with_null_migration_demes(self):
+        model = """
+time_units: generations
+demes:
+- name: a
+  epochs:
+  - {start_size: 1}
+- name: b
+  epochs:
+  - {start_size: 1}
+migrations:
+- demes:
+  rate: 0.01
+"""
+        with pytest.raises(ValueError, match="must have a non-null value"):
+            demes.loads(model)
+        model = """
+time_units: generations
+defaults:
+  migration:
+    demes: [a, b]
+demes:
+- name: a
+  epochs:
+  - {start_size: 1}
+- name: b
+  epochs:
+  - {start_size: 1}
+migrations:
+- demes:
+  rate: 0.01
+"""
+        with pytest.raises(ValueError, match="must have a non-null value"):
+            demes.loads(model)
+
+
+class TestMultiDocument:
+    @pytest.mark.parametrize("yaml_file", tests.example_files())
+    def test_load_all_single_document(self, yaml_file):
+        # Loading files with one document should work with the multi-doc API.
+        graphs = list(demes.load_all(yaml_file))
+        assert len(graphs) == 1
+        graph1 = graphs[0]
+        graph2 = demes.load(yaml_file)
+        graph1.assert_close(graph2)
+
+    @pytest.mark.parametrize("simplified", [True, False])
+    @pytest.mark.parametrize("graph1", tests.example_graphs())
+    def test_dump_all_single_document(self, graph1, simplified):
+        # A single documents saved with the multi-doc API should be loadable
+        # with the regular single-doc API.
+        with tempfile.TemporaryDirectory() as tmpdir:
+            tmpfile = pathlib.Path(tmpdir) / "temp.yaml"
+            demes.dump_all([graph1], tmpfile, simplified=simplified)
+            graph2 = demes.load(tmpfile)
+        graph1.assert_close(graph2)
+
+    @pytest.mark.parametrize("simplified", [True, False])
+    def test_round_trip_file(self, simplified):
+        graphs1 = tests.example_graphs()
+        assert len(graphs1) > 1
+        with tempfile.TemporaryDirectory() as tmpdir:
+            tmpfile = pathlib.Path(tmpdir) / "multidoc.yaml"
+            demes.dump_all(graphs1, tmpfile, simplified=simplified)
+            graphs2 = list(demes.load_all(tmpfile))
+        assert len(graphs1) == len(graphs2)
+        for g1, g2 in zip(graphs1, graphs2):
+            g1.assert_close(g2)
+
+    @pytest.mark.parametrize("simplified", [True, False])
+    def test_round_trip_stream(self, simplified):
+        graphs1 = tests.example_graphs()
+        assert len(graphs1) > 1
+        with tempfile.TemporaryDirectory() as tmpdir:
+            tmpfile = pathlib.Path(tmpdir) / "multidoc.yaml"
+            with open(tmpfile, "w") as f:
+                demes.dump_all(graphs1, f, simplified=simplified)
+            with open(tmpfile) as f:
+                graphs2 = list(demes.load_all(f))
+        assert len(graphs1) == len(graphs2)
+        for g1, g2 in zip(graphs1, graphs2):
+            g1.assert_close(g2)
+
+    @pytest.mark.parametrize("simplified", [True, False])
+    def test_round_trip_no_end_document_marker(self, simplified):
+        graphs1 = tests.example_graphs()
+        assert len(graphs1) > 1
+        with tempfile.TemporaryDirectory() as tmpdir:
+            tmpfile = pathlib.Path(tmpdir) / "multidoc.yaml"
+            with open(tmpfile, "w") as f:
+                for j, graph in enumerate(graphs1):
+                    if j > 0:
+                        # Output a start marker between documents.
+                        print("---", file=f)
+                    demes.dump(graph, f, simplified=simplified)
+            graphs2 = list(demes.load_all(tmpfile))
+        assert len(graphs1) == len(graphs2)
+        for g1, g2 in zip(graphs1, graphs2):
+            g1.assert_close(g2)
+
+    @pytest.mark.parametrize("simplified", [True, False])
+    def test_empty_file(self, simplified):
+        with tempfile.TemporaryDirectory() as tmpdir:
+            tmpfile = pathlib.Path(tmpdir) / "empty.yaml"
+            demes.dump_all([], tmpfile)
+            assert tmpfile.exists()
+            assert tmpfile.stat().st_size == 0
+            graphs = list(demes.load_all(tmpfile))
+            assert len(graphs) == 0
+
+
+class TestOpenFilePolymorph:
+    def test_fileobj_doesnt_get_closed_1(self):
+        devnull = open(os.devnull)
+        with demes.load_dump._open_file_polymorph(devnull, "w") as f:
+            pass
+        assert not f.closed
+        assert not devnull.closed
+        devnull.close()
+
+    def test_fileobj_doesnt_get_closed_2(self):
+        devnull = open(os.devnull)
+        try:
+            with demes.load_dump._open_file_polymorph(devnull, "w") as f:
+                raise ValueError
+        except ValueError:
+            pass
+        assert not f.closed
+        assert not devnull.closed
+        devnull.close()
+
+    def test_no_file_descriptor_leak_1(self):
+        with demes.load_dump._open_file_polymorph(os.devnull, "w") as f:
+            pass
+        assert f.closed
+
+    def test_no_file_descriptor_leak_2(self):
+        try:
+            with demes.load_dump._open_file_polymorph(os.devnull, "w") as f:
+                raise ValueError
+        except ValueError:
+            pass
+        assert f.closed
